@@ -176,6 +176,7 @@ struct tsm_vte {
 	unsigned int mouse_event;
 	unsigned int mouse_last_col;
 	unsigned int mouse_last_row;
+	bool bracketed_paste;
 
 	uint8_t (*custom_palette_storage)[3];
 	uint8_t (*palette)[3];
@@ -1756,6 +1757,9 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 			    continue;
 			}
 			continue;
+		case TSM_VTE_BRACKETED_PASTE:
+			vte->bracketed_paste = set;
+			continue;
 		default:
 			llog_debug(vte, "unknown DEC %set-Mode %d",
 				   set?"S":"Res", vte->csi_argv[i]);
@@ -1798,6 +1802,21 @@ static void csi_dsr(struct tsm_vte *vte)
 	}
 }
 
+static void csi_report_window_size(struct tsm_vte *vte)
+{
+	char buf[64];
+	unsigned int w, h, len, resp;
+
+	resp = vte->csi_argv[0] == 18 ? 8 : 9;
+
+	w = tsm_screen_get_width(vte->con);
+	h = tsm_screen_get_height(vte->con);
+	len = snprintf(buf, sizeof(buf), "\e[%u;%u;%ut", resp,  h + 1, w + 1);
+	if (len >= sizeof(buf))
+		return;
+	vte_write(vte, buf, len);
+}
+
 static void do_csi(struct tsm_vte *vte, uint32_t data)
 {
 	int num, x, y, upper, lower;
@@ -2042,6 +2061,16 @@ static void do_csi(struct tsm_vte *vte, uint32_t data)
 			num = 1;
 		tsm_screen_scroll_down(vte->con, num);
 		break;
+	case 't': /* ST */
+		if (vte->csi_argv[0] == 18 || vte->csi_argv[0] == 19)
+			csi_report_window_size(vte);
+		else
+			llog_debug(vte, "unhandled CSI t sequence %d", vte->csi_argv[0]);
+		break;
+	case 'b': /* Repeat last char */
+		num = vte->csi_argv[0];
+		tsm_screen_repeat_char(vte->con, num);
+		break;
 	default:
 		llog_debug(vte, "unhandled CSI sequence %c", data);
 	}
@@ -2088,7 +2117,7 @@ static void vte_write_xcolor(struct tsm_vte *vte, char *code,
 			     const char *end_seq,
 			     uint8_t r, uint8_t g, uint8_t b)
 {
-	char buf[32];
+	char buf[64];
 	snprintf(buf, sizeof(buf), "\e]%s;rgb:%02x%02x/%02x%02x/%02x%02x%s",
 		 code, r, r, g, g, b, b, end_seq);
 	vte_write(vte, buf, strlen(buf));
@@ -3246,7 +3275,13 @@ bool tsm_vte_handle_mouse(struct tsm_vte *vte, unsigned int cell_x,
 	bool pressed = event & TSM_MOUSE_EVENT_PRESSED;
 
 	/* drop move event if we don't wait for move events */
-	if ((vte->mouse_mode == TSM_VTE_MOUSE_MODE_X10 || vte->mouse_event != TSM_VTE_MOUSE_EVENT_ANY) && (event & TSM_MOUSE_EVENT_MOVED)) {
+	/* In mode 1002 (BTN), accept MOVED with button pressed (drag, button >= 32) */
+	/* In mode 1003 (ANY), accept all MOVED events */
+	bool is_drag = (button >= 32 && button <= 34);
+	if ((vte->mouse_mode == TSM_VTE_MOUSE_MODE_X10 ||
+	     (vte->mouse_event == TSM_VTE_MOUSE_EVENT_BTN && !is_drag) ||
+	     (vte->mouse_event != TSM_VTE_MOUSE_EVENT_BTN && vte->mouse_event != TSM_VTE_MOUSE_EVENT_ANY)) &&
+	    (event & TSM_MOUSE_EVENT_MOVED)) {
 		return false;
 	}
 
@@ -3290,13 +3325,18 @@ bool tsm_vte_handle_mouse(struct tsm_vte *vte, unsigned int cell_x,
 
 		vte_write(vte, buffer, strlen(buffer));
 		return true;
-	} else if (vte->mouse_mode == TSM_VTE_MOUSE_MODE_SGR && vte->mouse_event) {
+	} else if (vte->mouse_mode == TSM_VTE_MOUSE_MODE_SGR) {
 		if (event & TSM_MOUSE_EVENT_MOVED) {
 			if (cell_x == vte->mouse_last_col && cell_y == vte->mouse_last_row) {
 				return false;
 			}
 
-			reply_flags = 35;
+			/* For drags (button >= 32), use actual button value from reply_flags.
+			 * For hover motion (button < 32, only in mode ANY), use 35 (move marker). */
+			if (button < 32) {
+				reply_flags = 35;
+			}
+			/* else: reply_flags already set to button | modifiers for drags */
 			pressed = true;
 
 			vte->mouse_last_col = cell_x;
@@ -3307,9 +3347,14 @@ bool tsm_vte_handle_mouse(struct tsm_vte *vte, unsigned int cell_x,
 
 		vte_write(vte, buffer, strlen(buffer));
 		return true;
-	} else if (vte->mouse_mode == TSM_VTE_MOUSE_MODE_PIXEL && vte->mouse_event) {
+	} else if (vte->mouse_mode == TSM_VTE_MOUSE_MODE_PIXEL) {
 		if (event == TSM_MOUSE_EVENT_MOVED) {
-			reply_flags = 35;
+			/* For drags (button >= 32), use actual button value from reply_flags.
+			 * For hover motion (button < 32, only in mode ANY), use 35 (move marker). */
+			if (button < 32) {
+				reply_flags = 35;
+			}
+			/* else: reply_flags already set to button | modifiers for drags */
 			pressed = true;
 		}
 
@@ -3321,3 +3366,25 @@ bool tsm_vte_handle_mouse(struct tsm_vte *vte, unsigned int cell_x,
 
 	return false;
 }
+
+void tsm_vte_paste(struct tsm_vte *vte, const char *data)
+{
+	if (!vte || !data)
+		return;
+
+	if (vte->bracketed_paste) {
+		const char start[] = "\e[200~";
+		const char end[] = "\e[201~";
+		char *buf;
+		int len;
+
+		len = strlen(data) + sizeof(start) + sizeof(end) + 1;
+		buf = malloc(len);
+		if (!buf)
+			return;
+		snprintf(buf, len, "%s%s%s", start, data, end);
+		vte_write(vte, buf, strlen(buf));
+		free(buf);
+	} else
+		vte_write(vte, data, strlen(data));
+}
\ No newline at end of file
